'use strict';

/*
 *+------------------------------------------------------------------------+
 *| Licensed Materials - Property of IBM
 *| IBM Cognos Products: Smarts App
 *| (C) Copyright IBM Corp. 2020
 *|
 *| US Government Users Restricted Rights - Use, duplication or disclosure
 *| restricted by GSA ADP Schedule Contract with IBM Corp.
 *+------------------------------------------------------------------------+
 */
define(['jquery', 'dashboard-analytics/visualizations/renderer/VisView', './DataSet', './KeyDrivers', 'com/ibm/vida/control/vida', 'com/ibm/vida/utils/vida-utils'], function ($, Base, DataSet, KeyDrivers, vida, vidautils) {

	// defined messages to users
	var msg_title = 'What If Analysis';
	var msg_title_for = 'What If Analysis for {0}';
	var msg_create_model = 'Create model';
	var msg_train_model = 'Train model';
	var msg_training = 'Creating machine learning model. This process takes approximately 1 minute.';
	var msg_retraining = 'Applying additional training. This process takes approximately 1 minute.';
	var msg_autofill = 'Autofill input columns';
	var msg_filling = 'Please complete required fields in the fields tab.';
	var msg_few_points = 'Warning: specified columns do not contain sufficient data items. Please change your column selections or filters.';
	var msg_start_analysis = 'Click "Create model" to create a machine learning model and start your analysis.';
	var msg_model_accuracy = 'Prediction strength is {0}.';
	var msg_fetching_drivers = 'Autofilling Input Columns fields...';
	var msg_drivers_failed = 'Autofilling failed to complete. Please specify Input Columns manually.';
	var msg_drivers_succeeded = 'Found {0} input features';
	var msg_categorical = '(Categorical)';
	var msg_scenario_increase = '{0} +{1}';
	var msg_scenario_decrease = '{0} -{1}';
	var msg_and = ' and ';
	var msg_prediction_no_change = 'Prediction (no change)';

	var ATTR_INPUT_ID = 'data-input';

	var INITIAL_EPOCHS = 500;
	var CONTINUE_EPOCHS = 200;

	/*
  * Added a dependency to the vida-utils StringEx utility class since it provides useful format and join functions.
  * Dashboard/LiveWidget may provide an alternative utility that may be a better alternative, if you want to avoid
  * the dependency on Vida's utilities.
  */
	var StringEx = vidautils.StringEx;

	var bundles = Object.freeze([
	// Area
	{
		id: 'area',
		location: 'com/ibm/vida/vizbundles/rave2bundlearea'
	},
	// Bar
	{
		id: 'bar',
		location: 'com/ibm/vida/vizbundles/rave2bundlebar'
	},
	// Bubble
	{
		id: 'bubble',
		location: 'com/ibm/vida/vizbundles/rave2bubble'
	},
	// Composite
	{
		id: 'composite',
		location: 'com/ibm/vida/vizbundles/rave2bundlecomposite'
	},
	// Line
	{
		id: 'line',
		location: 'com/ibm/vida/vizbundles/rave2line'
	},
	// Pie
	{
		id: 'pie',
		location: 'com/ibm/vida/vizbundles/rave2bundlepie'
	}]);

	/**
  * VIPR acts as a static repository; web applications should only create a single instance.
  * The used vida widget (SmallMultipleGrid) will load the visualization bundle using this vipr instance.
  */
	var vipr = vida.VIPR.create({
		bundles: bundles
	});

	/**
  * Resolves tuple indices to a string value for the model.
  * @param {number} _value The value to resolve.
  * @param {Column} _column The Column to resolve the value for.
  */
	function valueForModel(_value, _column) {
		return _column.type === 'measure' ? _value : String(_value);
	}

	/**
  * Match non-context columns.
  * @param {Column} _column The column to check
  * @returns {boolean} `false` if the column has type `'context'`, `true` otherwise.
  */
	function nonContext(_column) {
		return _column.kind !== 'context';
	}

	/**
  * Create a percentage formatter.
  * @param {string|undefined} _locale The locale to use for the formatter.
  * @returns {Intl.NumberFormat}
  */
	function createPercentFormatter(_locale) {
		if (typeof Intl !== 'undefined' && typeof Intl.NumberFormat !== 'undefined') {
			return new Intl.NumberFormat(_locale, {
				style: 'percent',
				minimumFractionDigits: 0,
				maximumFractionDigits: 1
			});
		} else {
			return {
				format: function format(_value) {
					return Math.round(_value * 1000) / 10 + '%';
				}
			};
		}
	}

	var Widget = Base.extend({
		visModel: null,
		logger: null,
		ajaxSvc: null,
		whatIfServicePromise: null,
		/**
   * SourceTable instance containing the Dashboard query result.
   * This is used to create the training model, and as basis for the visualization data.
   */
		sourceTable: null,
		/**
   * DeltaTable contains a view to a predict result of the last executed scenario.
   * It can be processed into visualization data.
   */
		deltaTable: null,
		/**
   * ActiveScenario is the scenario that was added to represent the "live view" of changing the sliders.
   */
		activeScenario: null,
		/** Data for the Vida visualization. */
		viprData: null,
		/** AI model. */
		model: null,
		/** Identifier of the target column. */
		targetColumnId: null,
		/** Current prediction value. */
		prediction: null,
		config: null,
		/** The selected bundle id */
		_bundleId: null,
		/** A percentage forrmatter */
		_percentFmt: null,

		// Nodes and Widgets
		_widgetTitleNode: null,
		_menuNode: null,
		_trainButton: null,
		_inputButton: null,
		_widgetStatusNode: null,
		_inputsContainerNode: null,
		_inputsTBodyNode: null,
		_visualizationSectionNode: null,
		_vidaWidget: null,

		init: function init(_options) {
			Widget.inherited('init', this, arguments);

			Object.defineProperties(this, {
				visModel: {
					enumerable: true,
					value: _options.visModel
				},
				logger: {
					enumerable: true,
					value: _options.logger
				},
				ajaxSvc: {
					enumerable: true,
					value: _options.dashboardApi.getGlassCoreSvc('.Ajax')
				},
				whatIfServicePromise: {
					enumerable: true,
					value: _options.dashboardApi.getServiceWhenReady('.WhatIfService')
				},
				sourceTable: {
					enumerable: true,
					writable: true,
					value: null
				},
				deltaTable: {
					enumerable: true,
					writable: true,
					value: null
				},
				activeScenario: {
					enumerable: true,
					writable: true,
					value: null
				},
				viprData: {
					enumerable: true,
					writable: true,
					value: null
				},
				model: {
					enumerable: true,
					writable: true,
					value: null
				},
				targetColumnId: {
					enumerable: true,
					writable: true,
					value: null
				},
				prediction: {
					enumerable: true,
					writable: true,
					value: null
				},
				config: {
					enumerable: true,
					writable: true,
					value: null
				},
				_bundleId: {
					writable: true,
					value: null
				},
				// TODO: get preferred locale via API call?
				_percentFmt: {
					value: createPercentFormatter()
				}
			});

			this._createWidgetLayout();
		},

		_hasColumnInSlot: function _hasColumnInSlot(inputSlot, colId) {
			var inputList = inputSlot.getDataItemList();
			for (var i = 0; i < inputList.length; i++) {
				if (inputList[i].getColumnId() === colId) {
					return true;
				}
			}
			return false;
		},

		_showVisualization: function _showVisualization(_visible) {
			this._visualizationSectionNode.css('display', _visible ? '' : 'none');
		},

		/**
   * Creates the general and static parts of the widget, such as the header, inputs area etc.
   */
		_createWidgetLayout: function _createWidgetLayout() {
			// The top section consists of the title, a "menu" and a status message
			var topSectionNode = $('<div></div>').appendTo(this.$el);

			var widgetTitleNode = $('<h3>' + msg_title + '</h3>').appendTo(topSectionNode);

			var menuNode = $('<div class="menu"></div>').appendTo(topSectionNode);

			// Train button
			var trainButton = $('<button class="btn btn-default">Create model</button>').appendTo(menuNode);
			trainButton.css({
				'margin-bottom': '5px',
				'margin-right': '10px'
			});
			trainButton.click(this.train.bind(this));

			// Auto feature select button
			var inputButton = $('<button class="btn btn-default">' + msg_autofill + '</button>').appendTo(menuNode);
			inputButton.css({
				'margin-bottom': '5px',
				'margin-right': '10px'
			});
			inputButton.click(this.autoFeatureSelect.bind(this));

			// Status
			var statusContainerNode = $('<div class="well well-sm"></div>').appendTo(menuNode);
			var widgetStatusNode = $('<span>' + msg_filling + '</span>').appendTo(statusContainerNode);

			var headerHeight = '10em';

			// Inputs table (starts empty)
			var inputsContainerNode = $('<div style="position:absolute;left:0;top:' + headerHeight + ';width:45%;bottom:0;"></div>').appendTo(this.$el);
			var inputsTableNode = $('<table></table>').appendTo(inputsContainerNode);
			var inputsTBodyNode = $('<tbody></tbody>').appendTo(inputsTableNode);

			var visualizationSectionNode = $('<div style="position:absolute;top:' + headerHeight + ';right:0;width:50%;bottom:0;"></div>').appendTo(this.$el);
			// Visualization
			var vidaWidget = new vida.SmallMultipleGrid({
				vipr: vipr,
				style: 'width:100%;height:100%'
			});
			visualizationSectionNode.append(vidaWidget.domNode);

			Object.defineProperties(this, {
				_widgetTitleNode: {
					value: widgetTitleNode
				},
				_menuNode: {
					value: menuNode
				},
				_trainButton: {
					value: trainButton
				},
				_inputButton: {
					value: inputButton
				},
				_widgetStatusNode: {
					value: widgetStatusNode
				},
				_inputsContainerNode: {
					value: inputsContainerNode
				},
				_inputsTBodyNode: {
					value: inputsTBodyNode
				},
				_visualizationSectionNode: {
					value: visualizationSectionNode
				},
				_vidaWidget: {
					value: vidaWidget
				},
				_actionHandler: {
					// The ActionHandler provides default interaction such as zoom, and provides callbacks for other
					// actions such as tooltip.
					value: new vida.ActionHandler(vidaWidget, {
						hasTooltip: this._hasTooltip.bind(this),
						showTooltip: this._showTooltip.bind(this),
						updateTooltip: this._updateTooltip.bind(this),
						hideTooltip: this._hideTooltip.bind(this)
					})
				}
			});

			this._showVisualization(false);

			// input updates
			inputsTBodyNode.on('change', 'input[type="number"]', this._numberInputChange.bind(this));
			// sliders updates
			inputsTBodyNode.on('change', 'input[type="range"]', this._rangeInputChange.bind(this));
		},

		render: function render(renderInfo) {
			var _this = this;

			var renderVida = false;
			if (this.isMappingComplete()) {
				/*
     * Typically, users don not apply filters when training the model.
     * However, filters can be applied to the result (charts).
     * For example: in case the context slot shows a large number of values,
     * users should be able to filter the result, without re-training.
     * If data is refreshed, but the same columns presented (shape and order),
     * new SourceTable and DeltaTable are created.
     * */
				if (renderInfo.refresh.data || !this.sourceTable && renderInfo.data.getResult()) {
					var previousTable = this.sourceTable;
					this.sourceTable = DataSet.createTable(renderInfo.data.getResult(), this.visualization.getSlots());
					var sameShape = !!previousTable && this._sameShape(this.sourceTable, previousTable);

					// Create a DeltaTable from the sourceTable
					this.deltaTable = DataSet.createDeltaTable(this.sourceTable);
					this.activeScenario = null;

					var contextDataItemCount = this.sourceTable.columns.reduce(function (_acc, _column) {
						return _acc + (_column.kind === 'context' ? 1 : 0);
					}, 0);
					var inputDataItems = this.sourceTable.columns.filter(DataSet.isInputColumn);
					var targetDataItem = this.sourceTable.getColumn(-1);
					var rowCount = this.sourceTable.rowCount;
					var trainingData = {
						x: new Array(rowCount),
						y: new Array(rowCount)
					};
					// Populate training data
					for (var rowIndex = 0; rowIndex < rowCount; rowIndex++) {
						trainingData.x[rowIndex] = this.sourceTable.getRow(rowIndex, contextDataItemCount, -1, valueForModel);
						trainingData.y[rowIndex] = this.sourceTable.getValue(rowIndex, -1, valueForModel);
					}
					this.targetColumnId = targetDataItem.id;

					if (!sameShape) {
						this.model = null;
						this.viprData = null;
						renderVida = true;

						this._updateTrainButtonCaption();
						this.setStatus(rowCount < 100 ? msg_few_points : msg_start_analysis, 'danger');
						this.config = {
							epochs: INITIAL_EPOCHS,
							data: trainingData,
							types: inputDataItems.map(function (_column) {
								return _column.type;
							})
						};
						this._renderInputs(targetDataItem, inputDataItems);
						this.setEnabled(true);
					} else {
						this.predict();
					}
				} else if (this.viprData) {
					// If there is VIPR data, assume the visualization is visible and forward the render.
					renderVida = true;
				}
			} else {
				if (this.sourceTable) {
					this.sourceTable = null;
					this.deltaTable = null;
					this.viprData = null;
					this.activeScenario = null;
					this.config = null;
					this.model = null;
					this.targetColumnId = null;

					// Apply the "clear" state
					renderVida = true;

					this._updateTrainButtonCaption();
					// Clear the inputs table
					this._renderInputs(null, null);
				}
				this._setButtonsEnabled(false);
				var targetSlotItems = this.visualization.getSlots().getSlot('target').getDataItemList();
				if (targetSlotItems.length > 0) {
					this._inputButton.attr('disabled', null);
					// needed for autofilling rest call
					this.targetColumnId = targetSlotItems[0].getColumnId();
				}
				this.setWidgetTitle(targetSlotItems.length > 0 ? targetSlotItems[0].getLabel() : null);
			}

			// Call super.render
			var result = Widget.inherited('render', this, arguments);
			if (renderVida) {
				if (!this.viprData) {
					this._vidaWidget.setData(null);
				}
				// Invoke Vida's render as soon as the widget render is complete.
				result = result.then(function () {
					return _this._vidaWidget.render('client');
				});
				// If there is no data, hide the visualization after render completes.
				if (!this.viprData) {
					// Sadly this is necessary since most Vida bundles do not clear on setData(null),
					// instead keeping the last state.
					result = result.then(function () {
						return _this._showVisualization(false);
					});
				}
			}
			return result;
		},

		/**
   * Renders the target and input fields in a table.
   * There are 3 columns, one for the label, one for the slider and one for the value.
   */
		_renderInputs: function _renderInputs(_targetDataItem, _inputColumns) {
			this._inputsTBodyNode.empty();
			var tr = void 0;
			var td = void 0;

			/*
   if (_targetDataItem) {
   	// Target
   	tr = $('<tr/>').appendTo(this._inputsTBodyNode);
   		td = $('<td colspan="3" style="padding:5px;"/>').appendTo(tr);
   	td.append('<h4>Target</h4>');
   		// show the target variable name (col 1)
   	tr = $('<tr/>').appendTo(this._inputsTBodyNode);
   	td = $('<td/>').appendTo(tr);
   	td.append('<i class="glyphicon status"></i> ' + _targetDataItem.label).css('padding', '2px').css('padding-left', '5px');
   		// no slider for target (col 2)
   	$('<td colspan="2"/>').appendTo(tr);
   }
   */

			if (_inputColumns) {
				// show the input variables header
				tr = $('<tr/>').appendTo(this._inputsTBodyNode);
				td = $('<td colspan="3" style="padding:5px;"/>').appendTo(tr);
				td.append('<h4>Scenario (Input Columns % change)</h4>');

				// show the variables
				for (var i = 0; i < _inputColumns.length; i++) {
					// get data info
					var column = _inputColumns[i];
					tr = $('<tr/>').appendTo(this._inputsTBodyNode);

					// show the input variable name (col 1)
					td = $('<td/>').appendTo(tr);
					td.append('<i class="glyphicon glyphicon-transfer status"></i> ' + column.label).css('padding', '2px').css('padding-left', '5px');

					if (column.type === 'measure') {
						var input = void 0;
						// config values for sider
						// const domain = this.sourceTable.calculateDomain(column);

						td = $('<td/>').css('padding-left', '5px').css('padding-right', '10px').appendTo(tr);
						input = $('<input type="range" min="' + -100 + '" max="' + 100 + '" class="slider status">').appendTo(td);
						input.attr(ATTR_INPUT_ID, column.id);

						td = $('<td/>').css('padding-left', '5px').appendTo(tr);
						input = $('<input type="number" class="form-control status" style="padding:2px; height:22px" value="' + 0 + '">').appendTo(td);
						input.attr(ATTR_INPUT_ID, column.id);
					} else /*if(column.type === TYPE_CATEGORICAL)*/{
							/*
       // create all possible input options
       const options = column.tuples.map(function (_tuple, _index){
       	return '<option value=' + '"' + _index + '">' + _tuple.label + '</option>';
       });
       //Add no change option to the beginning of the array
       options.unshift('<option value="-1" style="font-style:italic;">' + msg_no_change + '</option>');
       	// drop down list
       td = $('<td colspan="2" style="padding: 10px 0"/>').appendTo(tr);
       let selector = $('<select id="value' + i + '" class="slider status" style="width:100%;" disabled/>').appendTo(td);
       selector.attr(ATTR_INPUT_ID, column.id);
       selector.append(options.join(''));
       */
							td = $('<td colspan="2" style="padding: 10px 0"/>').appendTo(tr);
							td.text(msg_categorical);
						}
				}
			}
		},

		/**
   * This function changes the rectangle of data by the variableValues value of each column.
   * the change for measure columns are in percentage. For cat columns, if the variableValues is null,
   * the values are kept as it, o.w. we change all values to the value in variableValues
   * For example, if rectangle of data is [ [10,10, '0'],[20,50, '1']]:
   * CASE1 : variableValues is [0,-50, '-1'] --> result: [[10,5, '0'],[20,25, '1']]
   * CASE1 : variableValues is [100,-50, '1'] --> result: [[20,5, '1'],[40,25, '1']]
   */
		changing_columns_by: function changing_columns_by(data, inputColumns, variableValues) {
			return data.map(function (row) {
				return row.map(function (_sourceValue, colIndex) {
					var delta = variableValues[colIndex];
					// If the delta value is `null`, the existing value should be used.
					if (delta === null) {
						return _sourceValue;
					}
					// For measures, adjust the values by the provided offset.
					// For categoricals, the delta should be the tuple index (number).
					return inputColumns[colIndex].type === 'measure' ? _sourceValue * delta : delta;
				});
			});
		},

		predict: function predict() {
			var _this2 = this;

			if (!this.model) {
				// Cannot predict without a model
				return;
			}
			var inputColumns = this.sourceTable.columns.filter(DataSet.isInputColumn);
			/** The scenarioNameParts are added as a side effect of the variableValues map function */
			var scenarioNameParts = [];
			var variableValues = inputColumns.map(function (_column) {
				// Simply persist categorical inputs by using the the source row value.
				if (_column.type === 'categorical') return null;
				// Build a selector to target the input field
				var attrSelector = '[' + ATTR_INPUT_ID + '="' + _column.id + '"]';
				var inputNode = this._inputsTBodyNode.find('input[type="number"]' + attrSelector + ',select' + attrSelector);
				// For measures, simply parse the value to a ratio.
				var delta = parseFloat(inputNode.val()) / 100;
				// Only include values that actually change.
				if (delta !== 0) {
					scenarioNameParts.push(StringEx.format(
					// Select the correct scenario message, increase or decrease
					delta < 0 ? msg_scenario_decrease : msg_scenario_increase,
					// Replacement 0: column name
					_column.label,
					// Replacement 1: the delta. Take the absolute, since we select an increase/decrease message
					this._percentFmt.format(Math.abs(delta))));
				}
				return 1 + delta;
			}, this);
			// New model inputs
			var newInputs = this.changing_columns_by(this.config.data.x, inputColumns, variableValues);
			var newResults = this.model.predict(newInputs);
			// Create a scenario description from the deltas
			var scenarioCaption = scenarioNameParts.length < 1 ? msg_prediction_no_change : StringEx.join(scenarioNameParts, ', ', msg_and);
			if (this.activeScenario) {
				this.activeScenario.update(scenarioCaption, newInputs, newResults);
			} else {
				this.activeScenario = this.deltaTable.addScenario(scenarioCaption, newInputs, newResults);
			}

			this._vidaWidget.newViz('bar').then(function (_bundle) {
				_this2._mapDataToViz(_bundle.vizDef);
			});
		},

		/**
   * Create a slot mapping for the visualization and create data for it.
   * @param {VizDef} _vizDef The Vida visualization definition.
   */
		_mapDataToViz: function _mapDataToViz(_vizDef) {
			var columns = this.deltaTable.columns;
			var mapping = DataSet.createColumnMapping(_vizDef);
			var targetIndex = columns.findIndex(DataSet.isTargetColumn);

			/*
    * For now, use a hard coded slot mapping for the Bar's slots:
    * categories (x-axis)
    * values (y-axis)
    * yStart (not used)
    * target (not used)
    * color (legend items)
    */

			// Note that if both the Scenario and virtual Row column are mapped, dat will never be aggregated.

			// Always map the scenario column to the color slot
			mapping.add('color', columns[0]);
			// Map the predicted target to the values slot
			mapping.add('values', columns[targetIndex]);
			// If context is not provided, aggreagte the data 

			// See if there is any categorical context column left to show as categories.
			var categoricalContextColumn = columns.find(function (_column) {
				return _column.type === 'categorical' && _column.kind === 'context';
			});
			// If not, map the virtual "rows" column so data is not aggregated.
			mapping.add('categories', categoricalContextColumn);

			this.viprData = DataSet.createViprData(this.deltaTable, mapping);

			// It may look odd that the data is passed twice; this first argument is the data, the second is the slot
			// mapping (which is also implemented by our viprData object).
			this._vidaWidget.setData( /*IVIPRData*/this.viprData, /*ISlotMapping*/this.viprData);
			this._showVisualization(true);
			return this._vidaWidget.render('client');
		},

		train: function train() {
			var _this3 = this;

			this.setEnabled(false);

			return this.whatIfServicePromise.then(function (_service) {
				// Either continue training, or create a new model
				if (_this3.model) {
					_this3.setStatus(msg_retraining, 'danger');
					return _service.continueTraining(_this3.model, _this3.config.data, {
						epochs: CONTINUE_EPOCHS
					});
				} else {
					_this3.setStatus(msg_training, 'danger');
					return _service.createModel(_this3.config);
				}
			}).then(function (_response) {
				_this3.model = _response.model;
				_this3._updateTrainButtonCaption();
				var accuracy = 1 - Math.sqrt(1 - Math.max(0, _response.r2));
				// TODO: add ', Standard Error is ±' + _response.mae.toFixed(2)
				_this3.setStatus(StringEx.format(msg_model_accuracy, _this3._percentFmt.format(accuracy)), 'success');
				_this3.setEnabled(true);
				// continue with a prediction
				_this3.predict();
			}).catch(function (_error) {
				_this3.logger.error(_error);
				_this3.setError('Something went wrong');
			});
		},

		autoFeatureSelect: function autoFeatureSelect() {
			var _this4 = this;

			this.setStatus(msg_fetching_drivers, 'danger');

			KeyDrivers.getKeyDrivers(this.ajaxSvc, this.visModel.getModule(), this.targetColumnId).then(function (drivers) {
				var inputSlot = _this4.visualization.getSlots().getSlot('input');
				var items = [];

				for (var i = 0; i < drivers.length; i++) {
					if (!_this4._hasColumnInSlot(inputSlot, drivers[i])) {
						items.push({
							columnId: drivers[i]
						});
					}
				}
				_this4.setStatus(StringEx.format(msg_drivers_succeeded, items.length), 'success');
				inputSlot.addDataItems(items);
			}).catch(function (_error) {
				_this4.logger.error(_error);
				_this4.setStatus(msg_drivers_failed, 'danger');
			});
		},

		_numberInputChange: function _numberInputChange(_event) {
			var node = _event.target;
			this.$el.find('input[type="range"][' + ATTR_INPUT_ID + '="' + node.getAttribute(ATTR_INPUT_ID) + '"]').val(node.value);
			this.predict();
		},

		_rangeInputChange: function _rangeInputChange(_event) {
			var node = _event.target;
			this.$el.find('input[type="number"][' + ATTR_INPUT_ID + '="' + node.getAttribute(ATTR_INPUT_ID) + '"]').val(node.valueAsNumber);
			this.predict();
		},

		setWidgetTitle: function setWidgetTitle(_target) {
			this._widgetTitleNode.text(_target ? StringEx.format(msg_title_for, _target) : msg_title);
		},


		setStatus: function setStatus(_message, _style) {
			this._widgetStatusNode.text(_message).toggleClass('text-muted', _style === 'muted').toggleClass('text-danger', _style === 'danger').toggleClass('text-success', _style === 'success');
		},

		_updateTrainButtonCaption: function _updateTrainButtonCaption() {
			this._trainButton.text(this.model ? msg_train_model : msg_create_model);
		},

		_setInputsEnabled: function _setInputsEnabled(_enabled) {
			var inputs = this._inputsTBodyNode.find('select[' + ATTR_INPUT_ID + '],input[' + ATTR_INPUT_ID + ']');
			inputs.attr('disabled', _enabled ? null : 'disabled');
		},

		_setButtonsEnabled: function _setButtonsEnabled(_enabled) {
			var inputs = this._menuNode.find('button');
			inputs.attr('disabled', _enabled ? null : 'disabled');
		},

		setEnabled: function setEnabled(_enabled) {
			this._setButtonsEnabled(_enabled);
			this._setInputsEnabled(_enabled && !!this.model);
		},

		setError: function setError(_message) {
			this.setStatus(_message, 'danger');
			this.setEnabled(true);
		},

		_hasTooltip: function _hasTooltip() {
			// TODO: implement tooltip
			return false;
		},

		_showTooltip: function _showTooltip(_point, _items) {
			// TODO: implement tooltip
			// Note: the _point is in local coordinates, relative to the visualization top-left.
			// To translate to client coordinates, use getBoundingClientRect() on the visualization node and add the
			// left, top values to the x, y values respectively.
			if (typeof console !== 'undefined') {
				console.debug('show tooltip for', _items);
			}
		},

		_updateTooltip: function _updateTooltip() /*_point, _items*/{
			// TODO: implement tooltip
		},

		_hideTooltip: function _hideTooltip() {
			// TODO: implement tooltip
		},
		/*
   * This function checks whether new data has same shape as old sourceTable (same set and 
  	order of columns), except for the context slot
   */
		_sameShape: function _sameShape(_sourceTable, _previousTable) {
			// Ignore context columns in comparison
			var columns = _sourceTable.columns.filter(nonContext);
			var previousColumns = _previousTable.columns.filter(nonContext);
			if (columns.length !== previousColumns.length) {
				return false;
			}
			for (var i = 0; i < columns.length; i++) {
				if (columns[i].id !== previousColumns[i].id || columns[i].type !== previousColumns[i].type) {
					return false;
				}
			}
			return true;
		}
	});

	return Widget;
});
//# sourceMappingURL=what_if.js.map